/**************************************************************
 copyright: (C)2007 by parafin
 email: operator@parafin.pp.ru

 This program is free software; you can redistribute  it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; either version 2 of
 the License, or (at your option) any later version.

 This program  is  distributed in  the  hope  that it  will  be
 useful, but  WITHOUT ANY  WARRANTY; without  even the  implied
 warranty  of  MERCHANTABILITY  or  FITNESS  FOR  A  PARTICULAR
 PURPOSE. See the GNU General Public License for more details.

 You should  have received  a copy  of the  GNU General  Public
 License along with  this program;  if not,  write to the  Free
 Software Foundation, Inc., 675 Mass Ave, Cambridge,  MA 02139,
 USA.
 **************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <mysql/mysql.h>
#include <ipt_ACCOUNT_cl.h>

/* Increase it if you want lengthier type names */
#define TYPE_LENGTH 20

char* server;
char* user;
char* password;
char* database;
char* table;
char* query;

const char* queryf    = "INSERT INTO %s (ip, bytes, packets, type, time,"
                        " direction) VALUES (?, ?, ?, ?, FROM_UNIXTIME(?), ?)";

const char* dir_in    = "in";
const char* dir_out   = "out";
#define DIR_LENGTH 3

MYSQL conn;
MYSQL_STMT* stmt_insert;
unsigned int curtime, ip, bytes, packets;
char type[TYPE_LENGTH+1], dir[DIR_LENGTH+1];
unsigned long type_length, dir_length;
struct ipt_ACCOUNT_context ctx;
int errors_occured = 0;

void sigint(int sig) {
	errors_occured = sig;
}

void terminate(int rc) {
	/* ipt_ACCOUNT deinitialization */
	ipt_ACCOUNT_deinit(&ctx);
	/* Begin MySQL deinitialization */
	if(stmt_insert) {
		if(mysql_stmt_close(stmt_insert)) {
			fprintf(stderr, "mysql_stmt_close failed: %s\n", mysql_stmt_error(stmt_insert));
		}
		stmt_insert = 0;
	}
	mysql_close(&conn);
	/* End of MySQL deinitialization */
	exit(rc);
}

void init() {
	/* Begin binding initialization */
	MYSQL_BIND bind[6];
	memset(bind, 0, sizeof(bind));
	bind[0].buffer_type   = MYSQL_TYPE_LONG;
	bind[0].is_unsigned   = 1;
	bind[0].buffer        = &ip;
	bind[1].buffer_type   = MYSQL_TYPE_LONG;
	bind[1].is_unsigned   = 1;
	bind[1].buffer        = &bytes;
	bind[2].buffer_type   = MYSQL_TYPE_LONG;
	bind[2].is_unsigned   = 1;
	bind[2].buffer        = &packets;
	bind[3].buffer_type   = MYSQL_TYPE_STRING;
	bind[3].buffer_length = TYPE_LENGTH+1;
	bind[3].length        = &type_length;
	bind[3].buffer        = type;
	bind[4].buffer_type   = MYSQL_TYPE_LONG;
	bind[4].is_unsigned   = 1;
	bind[4].buffer        = &curtime;
	bind[5].buffer_type   = MYSQL_TYPE_STRING;
	bind[5].buffer_length = DIR_LENGTH+1;
	bind[5].length        = &dir_length;
	bind[5].buffer        = dir;
	/* End if binding initialization */
	/* Begin MySQL initialization */
	mysql_init(&conn);
	if(!mysql_real_connect(&conn, server, user, password, database, 0, 0, 0)) {
		fprintf(stderr, "mysql_real_connect failed: %s\n", mysql_error(&conn));
		terminate(-1);
	}
	if(!(stmt_insert = mysql_stmt_init(&conn))) {
		fprintf(stderr, "mysql_stmt_init failed: %s\n", "out of memory");
		terminate(-1);
	}
	if(mysql_stmt_prepare(stmt_insert, query, strlen(query))) {
		fprintf(stderr, "mysql_stmt_prepare failed: %s\n", mysql_stmt_error(stmt_insert));
		terminate(-1);
	}
	if(mysql_stmt_bind_param(stmt_insert, bind)) {
		fprintf(stderr, "mysql_stmt_bind_param failed: %s\n", mysql_stmt_error(stmt_insert));
		terminate(-1);
	}
	/* End of MySQL initialization */
	/* ipt_ACCOUNT initialization */
	if(ipt_ACCOUNT_init(&ctx))
	{
		fprintf(stderr, "ipt_ACCOUNT_init failed: %s\n", ctx.error_str);
		terminate(-1);
	}
}

void real_work(const char* cur_table, const char* cur_type) {
	struct ipt_acc_handle_ip* entry;
	
	if(ipt_ACCOUNT_read_entries(&ctx, cur_table, 0)) {
		fprintf(stderr, "ipt_ACCOUNT_read_entries(%s) failed: %s\n", cur_table, ctx.error_str);
		terminate(-1);
	}
	curtime = time(0);
	while(entry = ipt_ACCOUNT_get_next_entry(&ctx)) {
		ip = htonl(entry->ip);
		bytes = entry->src_bytes;
		packets = entry->src_packets;
		strncpy(type, cur_type, TYPE_LENGTH+1);
		type_length = strlen(cur_type);
		strncpy(dir, dir_out, DIR_LENGTH+1);
		dir_length = strlen(dir_out);
		if(mysql_stmt_execute(stmt_insert) || (mysql_stmt_affected_rows(stmt_insert) != 1)) {
			fprintf(stderr, "mysql_stmt_execute(%s, %u, %s) failed: %s\n",
				cur_table, ip, dir, mysql_stmt_error(stmt_insert));
			printf("INSERT INTO %s (ip, bytes, packets, type, time, direction)"
				" VALUES (%u, %u, %u, '%s', FROM_UNIXTIME(%u), '%s')\n",
				table, ip, bytes, packets, type, curtime, dir);
			errors_occured++;
		}
		bytes = entry->dst_bytes;
		packets = entry->dst_packets;
		strncpy(dir, dir_in, DIR_LENGTH+1);
		dir_length = strlen(dir_in);
		if(mysql_stmt_execute(stmt_insert) || (mysql_stmt_affected_rows(stmt_insert) != 1)) {
			fprintf(stderr, "mysql_stmt_execute(%s, %u, %s) failed: %s\n",
				cur_table, ip, dir, mysql_stmt_error(stmt_insert));
			printf("INSERT INTO %s (ip, bytes, packets, type, time, direction)"
				" VALUES (%u, %u, %u, '%s', FROM_UNIXTIME(%u), '%s')\n",
				table, ip, bytes, packets, type, curtime, dir);
			errors_occured++;
		}
	}
}

int main(int argc, char** argv) {
	if(argc != 6) {
		fprintf(stderr, "Usage: %s server user password database table\n", argv[0]);
		exit(-1);
	}
	server = argv[1];
	user = argv[2];
	/* We must prevent password from showing up in `ps aux`, so we copy it and write NULLs to argv[3] */
	if(!(password = malloc(strlen(argv[3])+1))) {
		exit(-2);
	}
	strcpy(password, argv[3]);
	memset(argv[3], 0, strlen(argv[3]));
	database = argv[4];
	table = argv[5];
	if(!(query = malloc(strlen(queryf)+strlen(table)-1))) {
		exit(-2);
	}
	sprintf(query, queryf, table);
	signal(SIGINT, sigint);
	signal(SIGTERM, sigint);
	signal(SIGHUP, sigint);
	init();
	while(!errors_occured) {
#include "config.inc"
		if(!errors_occured) { sleep(60); }
	}
	terminate(errors_occured);
}
